@@ -28,7 +28,6 @@ def _get_config():
     parser = argparse.ArgumentParser(usage=PREFIX + _get_usage())
     parser.add_argument('-i', '--indent', type=int, default=2,
                         help='default: 2, range 0-8')
-    parser.add_argument('-z', '--compress', help='default: don\'t compress')
     parser.add_argument(
         '-f', '--fieldnames', action='store_true',
         help='if set the first row of csv file(s) is read as field names; '
@@ -67,7 +66,7 @@ def _postprocess_args(parser, config):
 
 
 def _postprocess_csv_args(parser, config):
-    if not config.file[-1].upper().endswith(DOT_UXF):
+    if not config.file[-1].upper().endswith((DOT_UXF, DOT_UXF_GZ)):
         parser.error('multiple infiles may only be converted to .uxf')
     for name in config.file[:-1]:
         if not name.upper().endswith(DOT_CSV):
@@ -82,10 +81,11 @@ def _postprocess_other_args(parser, config):
     config.outfile = config.file[-1]
     infile = config.infiles[0].upper()
     outfile = config.outfile.upper()
-    if infile.endswith(DOT_UXF) and outfile.endswith(DOT_UXF):
+    if (infile.endswith((DOT_UXF, DOT_UXF_GZ)) and
+            outfile.endswith((DOT_UXF, DOT_UXF_GZ))):
         parser.error('connot convert uxf to uxf, instead use: '
                      'python3 -m uxf infile.uxf outfile.uxf')
-    if outfile.endswith(DOT_UXF):
+    if outfile.endswith((DOT_UXF, DOT_UXF_GZ)):
         if infile.endswith(DOT_CSV):
             config.convert = csv_to_uxf
         elif infile.endswith(DOT_INI):
@@ -96,7 +96,7 @@ def _postprocess_other_args(parser, config):
             config.convert = sqlite_to_uxf
         elif infile.endswith(DOT_XML):
             config.convert = xml_to_uxf
-    elif infile.endswith(DOT_UXF):
+    elif infile.endswith((DOT_UXF, DOT_UXF_GZ)):
         if outfile.endswith(DOT_CSV):
             config.convert = uxf_to_csv
         elif outfile.endswith((DOT_JSN, DOT_JSON)):
@@ -108,7 +108,8 @@ def _postprocess_other_args(parser, config):
 
 
 def uxf_to_csv(config):
-    data, _ = uxf.load(config.infiles[0])
+    uxf_obj = uxf.load(config.infiles[0])
+    data = uxf_obj.data
     if isinstance(data, uxf.Table):
         with open(config.outfile, 'w') as file:
             writer = csv.writer(file, quoting=csv.QUOTE_NONNUMERIC)
@@ -129,12 +130,13 @@ def uxf_to_csv(config):
 
 
 def csv_to_uxf(config):
-    data, filename = _read_csv_to_data(config)
-    uxf.dump(config.outfile, data=data, custom=filename,
+    data, filename, ttypes = _read_csv_to_data(config)
+    uxf.dump(config.outfile, uxf.Uxf(data, filename, ttypes=ttypes),
              one_way_conversion=True)
 
 
 def _read_csv_to_data(config):
+    ttypes = {}
     data = None
     filename = config.infiles[0]
     with open(filename) as file:
@@ -142,14 +144,10 @@ def _read_csv_to_data(config):
         for row in reader:
             if data is None:
                 if config.fieldnames:
-                    name = pathlib.Path(filename).stem
-                    if not name[0].isupper():
-                        if name[0].isalpha():
-                            name = name[0].upper() + name[1:]
-                        else:
-                            name = f'T_{name}'
+                    name = uxf.canonicalize(pathlib.Path(filename).stem)
                     data = uxf.Table(name=name, fields=[uxf.Field(name)
                                      for name in row])
+                    ttypes[name] = data.ttype
                     continue
                 else:
                     data = []
@@ -158,24 +156,36 @@ def _read_csv_to_data(config):
                 data += row
             else:
                 data.append(row)
-    return data, filename
+    return data, filename, ttypes
 
 
 def multi_csv_to_uxf(config):
     infiles = config.infiles
     data = []
+    ttypes = {}
     for infile in infiles:
         config.infiles = [infile]
-        datum, _ = _read_csv_to_data(config)
+        datum, _, new_ttypes = _read_csv_to_data(config)
         data.append(datum)
-    uxf.dump(config.outfile, data=data, custom=' '.join(infiles),
+        if new_ttypes:
+            ttypes.update(new_ttypes)
+    uxf.dump(config.outfile,
+             uxf.Uxf(data, ' '.join(infiles), ttypes=ttypes),
              one_way_conversion=True)
 
 
 def uxf_to_json(config):
-    data, _ = uxf.load(config.infiles[0])
+    uxf_obj = uxf.load(config.infiles[0])
+    d = {}
+    if uxf_obj.custom is not None:
+        d[JSON_CUSTOM] = uxf_obj.custom
+    if uxf_obj.comment is not None:
+        d[JSON_COMMENT] = uxf_obj.comment
+    if uxf_obj.ttypes:
+        d[JSON_TTYPES] = list(uxf_obj.ttypes.values())
+    d[JSON_DATA] = uxf_obj.data
     with open(config.outfile, 'wt', encoding=UTF8) as file:
-        json.dump(data, file, cls=_JsonEncoder, indent=2)
+        json.dump(d, file, cls=_JsonEncoder, indent=2)
 
 
 class _JsonEncoder(json.JSONEncoder):
@@ -203,6 +213,10 @@ class _JsonEncoder(json.JSONEncoder):
                 name=obj.name, comment=obj.comment,
                 fields={field.name: field.vtype for field in obj.fields},
                 records=obj.records)}
+        if isinstance(obj, uxf.TType):
+            return {JSON_TTYPE: dict(
+                name=obj.name,
+                fields={field.name: field.vtype for field in obj.fields})}
         return json.JSONEncoder.default(self, obj)
 
 
@@ -240,8 +254,18 @@ def _json_encode_map(obj):
 def json_to_uxf(config):
     filename = config.infiles[0]
     with open(filename, 'rt', encoding=UTF8) as file:
-        data = json.load(file, object_hook=_json_naturalize)
-    uxf.dump(config.outfile, data=data, custom=filename)
+        d = json.load(file, object_hook=_json_naturalize)
+    custom = d.get(JSON_CUSTOM)
+    comment = d.get(JSON_COMMENT)
+    ttypes = None
+    ttype_list = d.get(JSON_TTYPES)
+    if ttype_list:
+        ttypes = {}
+        for ttype in ttype_list:
+            ttypes[ttype.name] = ttype
+    data = d.get(JSON_DATA)
+    uxf.dump(config.outfile, uxf.Uxf(data, custom, ttypes=ttypes,
+                                     comment=comment))
 
 
 def _json_naturalize(d):
@@ -274,12 +298,16 @@ def _json_naturalize(d):
                 key = key if itype is None else uxf.naturalize(key)
             m[key] = value
         return m
-    elif JSON_TABLE in d:
+    if JSON_TABLE in d:
         jtable = d[JSON_TABLE]
         fields = [uxf.Field(name, vtype) for name, vtype in
                   jtable[FIELDS].items()]
         return uxf.Table(name=jtable[NAME], comment=jtable[COMMENT],
                          fields=fields, records=jtable[RECORDS])
+    if JSON_TTYPE in d:
+        fields = [uxf.Field(name, vtype) for name, vtype in
+                  d[JSON_TTYPE][FIELDS].items()]
+        return uxf.TType(d[JSON_TTYPE][NAME], fields)
     return d
 
 
@@ -294,7 +322,7 @@ def ini_to_uxf(config):
             m = data[section] = uxf.Map()
             for key, value in d.items():
                 m[uxf.naturalize(key)] = uxf.naturalize(value)
-    uxf.dump(config.outfile, data=data, custom=filename,
+    uxf.dump(config.outfile, uxf.Uxf(data, filename),
              one_way_conversion=True)
 
 
@@ -327,7 +355,7 @@ def _uxf_to_sqlite(config, tables):
 
 
 def _create_table(db, table, table_index):
-    table_name = uxf._canonicalize(table.name, f'Table{table_index}')
+    table_name = uxf.canonicalize(table.name, f'T_{table_index}')
     sql = ['CREATE TABLE IF NOT EXISTS ', table_name, ' (']
     types = ['TEXT'] * len(table.fieldnames)
     if table.records:
@@ -340,7 +368,7 @@ def _create_table(db, table, table_index):
                 types[i] = 'BLOB'
     sep = ''
     for i, name in enumerate(table.fieldnames):
-        field_name = uxf._canonicalize(name, f'Field{i + 1}')
+        field_name = uxf.canonicalize(name, f'F_{i + 1}')
         sql += [sep, field_name, ' ', types[i]]
         sep = ', '
     sql += [');']
@@ -376,10 +404,16 @@ DOT_JSN = '.JSN'
 DOT_JSON = '.JSON'
 DOT_SQLITE = '.SQLITE'
 DOT_UXF = '.UXF'
+DOT_UXF_GZ = '.UXF.GZ'
 DOT_XML = '.XML'
 FIELDS = 'fields'
 ITYPE = 'itype'
 ITYPES = 'itypes'
+JSON_CUSTOM = 'UXF^custom'
+JSON_COMMENT = 'UXF^comment'
+JSON_TTYPES = 'UXF^ttypes'
+JSON_TTYPE = 'UXF^ttype'
+JSON_DATA = 'UXF^data'
 JSON_BYTES = 'UXF^bytes'
 JSON_DATE = 'UXF^date'
 JSON_DATETIME = 'UXF^datetime'
@@ -396,17 +430,17 @@ UXF = 'uxf'
 VTYPE = 'vtype'
 
 PREFIX = '''
-uxfconvert.py <infile.uxf> <outfile.{csv,json,sqlite,xml}>
-uxfconvert.py [-z|--compress] [-i|--indent=N] [-f|--fieldnames] \
-<infile.{csv,ini,json,sqlite,xml}> <outfile.uxf>
-uxfconvert.py [-z|--compress] [-i|--indent=N] [-f|--fieldnames] \
-<infile1.csv> [infile2.csv ... infileM.csv] <outfile.uxf>
+uxfconvert.py <infile.uxf[.gz]> <outfile.{csv,json,sqlite,xml}>
+uxfconvert.py [-i|--indent=N] [-f|--fieldnames] \
+<infile.{csv,ini,json,sqlite,xml}> <outfile.uxf[.gz]>
+uxfconvert.py [-i|--indent=N] [-f|--fieldnames] \
+<infile1.csv> [infile2.csv ... infileM.csv] <outfile.uxf[.gz]>
 
 '''
 
 USAGE = '''Converts to and from uxf format.
 
-If compress is set and the output is uxf, the uxf will be gzip compressed.
+If the outfile is .uxf.gz, the output will be gzip compressed UXF.
 
 Indent defaults to 2 (uxf's default); but can be set to any value 0-8.
 
@@ -414,7 +448,7 @@ If fieldnames is set and the infile(s) is(are) csv the first row of each
 infile will be read as field (column) names; otherwise all rows will be
 assumed to contain values.
 
-To produce compact uxf output use options: -z -i0.
+To produce compact uxf output use a .gz suffix and -i0.
 
 Converting from multiple csv files to uxf, the outfile will either be a
 list of tables (if the fieldnames option is given), or a list of lists of
@@ -428,9 +462,10 @@ comments are dropped). In a real application (e.g., migrating from ini to
 uxf), a custom ini parser would be needed. Converting uxf to ini is not
 supported.
 
-Converting sqlite to uxf only converts the sql tables and so won't
-roundtrip. Converting uxf to sqlite is only supported if the uxf file is a
-single scalar table or a list of scalar tables.
+Converting sqlite to uxf is purely for example purposes and only converts
+the sql tables and so won't roundtrip. Converting uxf to sqlite is again
+only for example purposes and only supported if the uxf file is a single
+scalar table or a list of scalar tables.
 
 Converting from uxf to json and back (i.e., using uxfconvert.py's own json
 format) roundtrips with perfect fidelity.
@@ -441,7 +476,7 @@ format) roundtrips with perfect fidelity.
 Support for uxf to uxf conversions is provided by the uxf.py module itself,
 which can be run directly or via python, e.g., `uxf.py infile.uxf
 outfile.uxf` or `python3 -m uxf infile.uxf outfile.uxf` with the same indent
-and compress options as here, plus additional options (use uxf.py's -h or
+and compression options as here, plus additional options (use uxf.py's -h or
 --help for details).'''
 
 
